#!/usr/bin/perl

# /usr/local/bin/matrixctl.pl
# Dieses Programm überwacht eine Änderungsdatei vom Web-Server und startet bei Bedarf den Matrixuhr-Dienst neu.
# Außerdem wird hier (falls vorhanden) der 1-Wire-Sensor abgefragt und dessen Wert alle 60 Sekunden gespeichert.
# Falls der Empfang der Wetterdaten nicht über MQTT erfolgt, wird zusätzlich der Server für die Wetterdaten abgefragt
# und das empfangene Datenpaket gespeichert.
# Scott-Falk Hühn, 10.07.2022

use strict;
use warnings;
use LWP::Simple;              # wird zur Wetterdatenabfrage benötigt

my $sendir = "/mnt/ram";      # Verzeichnis für flüchtige Sensor- und Alarmdaten
my $cfgdir = "/srv/sensor";   # Verzeichnis der Konfigurationsdateien
my $senif = "/dev/serial0";   # Serielle Schnittstelle des Raspberry Pi
my $cntmax = 12;              # Endwert für Zykluszähler in Schritten von 5 Sekunden (60 Sekunden)
my $SER;                      # Dateihandle der seriellen Schnittstelle
my $CFG;                      # Dateihandle einer Konfigurationsdatei
my $CHFILE;                   # Dateihandle für Change-Datei
my $W1FILE;                   # Dateihandle für 1-Wire-Dateien
my $WEFILE;                   # Dateihandle für Wetterdatei
my $chmode;                   # übermitteltes Change-Kommando
my $test;                     # Flag für Test-Modus, wird durch den Parameter "t" oder "T" beim Start aktiviert
my $count;                    # Zykluszähler
my $temp;                     # Zwischenspeicher
my $w1_num;                   # Anzahl erkannter 1-Wire-Sensoren
my $w1_sid;                   # Sensor-ID des ersten 1-Wire-Sensors
my $w1_crc;                   # gelesenes 1-Wire-Scratchpad und CRC-Prüfung
my $w1_val;                   # gelesener 1-Wire-Temperaturwert
my $w1_off;                   # gelesener 1-Wire-Offset (Korrekturwert)
my $w1_tmp;                   # korrigierter Temperaturwert
my $wename = "weatmu";        # Name der Wetter-Konfigurationsdatei
my $jsname = "weather.json";  # Name der Wetterdatei
my $we_bas = "https://api.openweathermap.org/data/2.5/weather?"; # Basis-URL für Wetterdatenabfrage
my $we_opt = "&units=metric&lang=de"; # weitere Optionen für die Wetterdatenabfrage
my $we_use;                   # Verwendung der Wetterdatenabfrage, wird aus Konfigurationsdatei gelesen (1 = ein)
my $we_aid;                   # APPID für OpenWeatherMap.org, wird aus Konfigurationsdatei gelesen
my $we_loc;                   # Name des Ortes für die Wetterdatenabfrage, wird aus Konfigurationsdatei gelesen
my $we_lat;                   # Breitengrad, falls kein Ortsname angegeben, wird aus Konfigurationsdatei gelesen
my $we_lon;                   # Längengrad, falls kein Ortsname angegeben, wird aus Konfigurationsdatei gelesen
my $we_url;                   # zusammengesetzte URL für die Wetterdatenabfage
my $we_res;                   # empfangenes Datenpaket mit den Wetterdaten im JSON-Format
my $sdname = "sdatmu";        # Name der MQTT-Konfigurationsdatei für spezielle Daten
my $sdtop1;                   # MQTT Topic für Wetterdaten

$test = $ARGV[0];                                             # Paramter 1 holen
if (defined $test) {                                          # wenn Parameter 1 angegeben
  if (($test eq 't') || ($test eq 'T')) { $test = 1; }        # wenn Zeichen 'l' oder 'L' -> Test-Modus einschalten
  else { $test = 0; }                                         # sonst Test-Modus ausschalten
}

$w1_num = 0;                                                  # Anzahl der 1-Wire-Sensoren = 0 setzen
if (open($W1FILE, "</sys/bus/w1/devices/w1_bus_master1/w1_master_slave_count")) { # Systemdatei mit Sensoranzahl öffnen
  $w1_num = <$W1FILE>;                                        # Anzahl der Sensoren lesen
  close($W1FILE);                                             # Datei schließen
  if ($w1_num ne '') { chomp($w1_num); }                      # wenn Daten vorhanden -> Zeilenendezeichen entfernen
}
if ($w1_num) {                                                # wenn Sensoren vorhanden
  if (open($W1FILE, "</sys/bus/w1/devices/w1_bus_master1/w1_master_slaves")) { # Systemdatei mit Sensorliste öffnen
    $w1_sid = <$W1FILE>;                                      # Sensor-ID des ersten Sensors lesen
    close($W1FILE);                                           # Datei schließen
    if ($w1_sid ne '') {                                      # wenn Daten vorhanden
      chomp($w1_sid);                                         # Zeilenendezeichen entfernen
      if (length($w1_sid) != 15) { undef $w1_sid; }           # wenn Sensor-ID nicht 15 Zeichen lang -> verwerfen
    }
  }
}                                                             # wenn Verzeichnis für flüchtige Daten nicht existiert
if (!(-d $sendir)) { $sendir = $cfgdir; }                     # Verzeichnis für Konfigurationsdaten verwenden
if ($test) {                                                  # wenn Test-Modus aktiv
  print "Speicherort für Daten: $sendir\n";                   # gewählten Speicherort ausgeben
}

if ($test) {                                                  # wenn Test-Modus aktiv
  print "1-Wire-Sensor: ";
  if (defined $w1_sid) { print "$w1_sid\n"; }                 # wenn Sensor-ID vorhanden -> Sensor-ID ausgeben
  else { print "nicht vorhanden\n"; }                         # wenn keine Sensor-ID vorhanden -> Hinweis ausgeben
}

$count = $cntmax;                                             # Zykluszähler auf Endwert setzen
while(1) {                                                    # Endlosschleife
  if (-e "$cfgdir/changemu") {                                # wenn Change-Datei vorhanden
    open($CHFILE, "<$cfgdir/changemu");                       # Change-Datei öffnen
    $chmode = <$CHFILE>;                                      # Change-Kommando lesen
    close($CHFILE);                                           # Change-Datei schließen
    unlink "$cfgdir/changemu";                                # Change-Datei löschen
    if (-e "$sendir/$jsname") { unlink "$sendir/$jsname"; }   # Wetterdatei löschen
    if (defined $chmode) {                                    # wenn Change-Kommando vorhanden
      chomp($chmode);                                         # Zeilenendezeichen löschen
      if ($chmode =~ /^Restart$/) {                           # wenn Kommando "Restart"
        if ($test) { print "Neustart Matrixuhr-Dienst\n"; }   # Test-Daten ausgeben
        system("systemctl restart matrixuhr");                # Matrixuhr-Dienst neu starten
        $count = $cntmax;                                     # Zykluszähler auf Endwert setzen
      }                                                       # (1-Wire und Wetter sofort ausführen)
    }
  }
  sleep 5;                                                    # 5 Sekunden warten
  $count ++;                                                  # Zykluszähler erhöhen
  if ($count >= $cntmax) {                                    # wenn Endwert erreicht
    $count = 0;                                               # Zykluszähler zurücksetzen

                                                              # 1-Wire-Sensorabfrage, einige Programmteile stammen von:
                                                              # http://www.netzmafia.de/skripten/hardware/RasPi
                                                              # /Projekt-Onewire/index.html

    undef $w1_val;                                            # zunächst 1-Wire-Temperaturwert löschen
    if (defined $w1_sid) {                                    # wenn 1-Wire-Sensor-ID vorhanden
      if (open($W1FILE, "</sys/bus/w1/devices/$w1_sid/w1_slave")) { # wenn Systemdatei für Sensorabfrage vorhanden
        $w1_crc = <$W1FILE>;                                  # erste Zeile lesen (Scratchpad und CRC)
        $temp = <$W1FILE>;                                    # zweite Zeile lesen (Scratchpad und Temperaturwert)
        close($W1FILE);                                       # Datei schließen
        if ($w1_crc =~ m/YES/) {                              # wenn CRC in Ordnung
          if ($temp ne '') {                                  # wenn Temperaturwert vorhanden
            $temp =~ /t=(\D*\d+)/i;                           # Temperaturwert lesen
            $w1_val = $1 / 1000;                              # Temperaturwert in °C umrechnen
            $w1_off = 0;                                      # 1-Wire-Offset auf 0 voreinstellen
            if (open($W1FILE, "<$cfgdir/s32mu")) {            # wenn 1-Wire-Offset-Datei vorhanden -> öffnen
              $temp = <$W1FILE>;                              # 1-Wire-Offset lesen
              close($W1FILE);                                 # 1-Wire-Offset-Datei schließen
            }
            if (defined $temp) {                              # wenn 1-Wire-Offset vorhanden
              chomp($temp);                                   # Zeilenendezeichen entfernen
              if ($temp =~ /^\-?\d*\.?\d*$/) {                # wenn 1-Wire-Offset plausibel
                if (($temp >= -5) && ($temp <= 5)) {          # wenn 1-Wire-Offset innerhalb des zulässigen Bereichs
                  $w1_off = $temp;                            # 1-Wire-Offset übernehmen
                }
              }
            }
            $w1_tmp = sprintf("%.1f", $w1_val + $w1_off);     # korrigierte Temperaturwert mit einer Nachkommastelle
            if (($w1_tmp == 0) && ($w1_tmp =~ /\-+/)) {       # wenn Spezialfall -0 eintritt
              $w1_tmp = sprintf("%.1f", 0);                   # als 0 mit einer Nachkommastelle speichern
            }
            open($W1FILE, ">$sendir/s32");                    # Sensordatei anlegen
            print $W1FILE "$w1_tmp\n";                        # Sensorwert schreiben
            close($W1FILE);                                   # Sensordatei schließen
            if ($test) {                                      # wenn Test-Modus aktiv
              $w1_val = sprintf("%.1f", $w1_val);             # Temperaturwert mit einer Nachkommastelle
              if (($w1_val == 0) && ($w1_val =~ /\-+/)) {     # wenn Spezialfall -0 eintritt
                $w1_val = sprintf("%.1f", 0);                 # als 0 mit einer Nachkommastelle speichern
              }
              print "Sensorwert: $w1_val°C\n";                # Sensorwert ausgeben (Log)
              print "Korrigiert: $w1_tmp°C\n";                # korrigierten Wert ausgeben (log)
            }
          }
        }
      }
    }
                                                              # Wetter-Konfiguration einlesen
    undef $we_use;                                            # Aktivierung löschen
    undef $we_aid;                                            # APPID löschen
    undef $we_loc;                                            # Ortsname löschen
    undef $we_lat;                                            # Breitengrad löschen
    undef $we_lon;                                            # Längengrad löschen
    if (open($CFG, "<$cfgdir/$wename")) {                     # wenn Wetter-Konfigurationsdatei vorhanden -> öffnen
      $we_use = <$CFG>;                                       # Aktivierung lesen
      $we_aid = <$CFG>;                                       # APPID lesen
      $we_loc = <$CFG>;                                       # Ortsname lesen
      $we_lat = <$CFG>;                                       # Breitengrad lesen
      $we_lon = <$CFG>;                                       # Längengrad lesen
      close($CFG);                                            # Wetter-Konfigurationsdatei schließen
    }
    if ((defined $we_use) && (defined $we_aid) && (defined $we_loc) &&
        (defined $we_lat) && (defined $we_lon)) {             # wenn alle erforderlichen Parameter vorhanden
      chomp($we_use);                                         # Zeilenendezeichen der Aktivierung entfernen
      chomp($we_aid);                                         # Zeilenendezeichen der APPID entfernen
      chomp($we_loc);                                         # Zeilenendezeichen des Ortsnamens entfernen
      chomp($we_lat);                                         # Zeilenendezeichen des Breitengrades entfernen
      chomp($we_lon);                                         # Zeilenendezeichen des Längengrades entfernen
      if (open($CFG, "<$cfgdir/$sdname")) {                   # wenn spezielle MQTT-Konfigurationsdatei vorh. -> öffnen
        $sdtop1 = <$CFG>;                                     # MQTT Topic für Wetterdaten lesen
        chomp($sdtop1);                                       # Zeilenendezeichen entfernen
        close($CFG);                                          # Wetter-Konfigurationsdatei schließen
      }
      else { $sdtop1 = ''; }                                  # wenn spezielle MQTT-Konfig. fehlt -> leeres MQTT Topic
      if ($sdtop1 ne '') { $we_use = 0; }                     # wenn Wetterdaten über MQTT -> Wetterabfrage deaktiv.
      if ($we_use) {                                          # wenn Wetterabfrage eingeschaltet
        if ($we_loc ne '') {                                  # wenn Ortsname vorhanden
          $we_url = $we_bas . "q=" . $we_loc . "&APPID=" . $we_aid . $we_opt;
        }                                                     # Wetter-URL für Abfrage über Ortsname zusammenstellen
        else {                                                # wenn kein Ortsname vorhanden
          $we_url = $we_bas . "lat=" . $we_lat . "&lon=" . $we_lon . "&APPID=" . $we_aid . $we_opt;
        }                                                     # Wetter-URL für Abfrage über Koordinaten zusammenstellen
        if ($test) { print "$we_url\n"; }                     # wenn Test-Modus aktiv -> Wetter-URL ausgeben
        $we_res = get($we_url);                               # Wetterdaten abfragen
        if (defined $we_res) {                                # wenn ein Datenpaket empfangen wurde
          if ((substr($we_res, 0, 1) eq '{') && (substr($we_res, -1, 1) eq '}')) { # wenn JSON-Objekt ({} vorhanden)
            open($WEFILE, ">$sendir/$jsname");                # Wetterdatei anlegen
            print $WEFILE "$we_res\n";                        # Wetterdaten schreiben
            close($WEFILE);                                   # Wetterdatei schließen
            if ($test) { print "$we_res\n"; }                 # wenn Log aktiv -> empfangenes Datenpaket ausgeben
          }
        }
      }
    }
  }
}
